import click

from torrra._version import __version__


@click.group(invoke_without_command=True)
@click.version_option(version=__version__, prog_name="torrra")
@click.option("--no-cache", is_flag=True, help="Disable caching mechanism.")
@click.pass_context
def cli(ctx: click.Context, no_cache: bool) -> None:
    if ctx.invoked_subcommand is None:
        from torrra.utils.indexer import run_with_default_indexer

        # detect indexer from config and execute
        run_with_default_indexer(no_cache=no_cache)


# --------------------------------------------------
# SEARCH
# --------------------------------------------------
@cli.command(help="Search for a torrent.")
@click.argument("search_query")
@click.option("--no-cache", is_flag=True, help="Disable caching mechanism.")
def search(search_query: str, no_cache: bool) -> None:
    from torrra.utils.indexer import run_with_default_indexer

    # detect indexer from config and execute with given search query
    run_with_default_indexer(no_cache=no_cache, search_query=search_query)


# --------------------------------------------------
# INDEXERS
# --------------------------------------------------
@cli.command(help="Use Jackett as the indexer.")
@click.option("--url", required=False, help="Jackett server URL.")
@click.option("--api-key", required=False, help="Jackett API key.")
@click.option("--no-cache", is_flag=True, help="Disable caching mechanism.")
def jackett(url: str | None, api_key: str | None, no_cache: bool) -> None:
    from torrra.utils.indexer import run_with_indexer

    # run app with jackett indexer
    run_with_indexer(
        name="jackett",
        indexer_cls_str="torrra.indexers.jackett.JackettIndexer",
        url=url,
        api_key=api_key,
        no_cache=no_cache,
    )


@cli.command(help="Use Prowlarr as the indexer.")
@click.option("--url", required=False, help="Prowlarr server URL.")
@click.option("--api-key", required=False, help="Prowlarr API key.")
@click.option("--no-cache", is_flag=True, help="Disable caching mechanism.")
def prowlarr(url: str | None, api_key: str | None, no_cache: bool):
    from torrra.utils.indexer import run_with_indexer

    # run app with prowlarr indexer
    run_with_indexer(
        name="prowlarr",
        indexer_cls_str="torrra.indexers.prowlarr.ProwlarrIndexer",
        url=url,
        api_key=api_key,
        no_cache=no_cache,
    )


# --------------------------------------------------
# CONFIG
# --------------------------------------------------
@cli.group(help="Manage configuration.")
def config():
    pass


@config.command(name="get", help="Get a config value.")
@click.argument("key")
def config_get(key: str):
    from torrra.core.config import get_config
    from torrra.core.exceptions import ConfigError

    try:
        click.echo(get_config().get(key))
    except ConfigError as e:
        click.secho(e, fg="red", err=True)


@config.command(name="set", help="Set a config value to path.")
@click.argument("key")
@click.argument("value")
def config_set(key: str, value: str):
    from torrra.core.config import get_config
    from torrra.core.exceptions import ConfigError

    try:
        get_config().set(key, value)
    except ConfigError as e:
        click.secho(e, fg="red", err=True)


@config.command(name="list", help="List all configurations.")
def config_list():
    from torrra.core.config import get_config
    from torrra.core.exceptions import ConfigError

    try:
        click.echo("\n".join(get_config().list()))
    except ConfigError as e:
        click.secho(e, fg="red", err=True)


if __name__ == "__main__":
    cli()



from typing import ClassVar

from textual import work
from textual.app import App
from textual.binding import Binding, BindingType
from textual.reactive import Reactive
from textual.types import CSSPathType

from torrra._types import Indexer
from torrra.core.config import get_config
from torrra.screens.home import HomeScreen
from torrra.screens.theme_selector import ThemeSelectorScreen
from torrra.screens.welcome import WelcomeScreen
from torrra.utils.fs import get_resource_path


class TorrraApp(App[None]):
    theme: Reactive[str]

    TITLE: str | None = "torrra"
    CSS_PATH: ClassVar[CSSPathType | None] = get_resource_path("app.tcss")
    ENABLE_COMMAND_PALETTE: ClassVar[bool] = False
    BINDINGS: ClassVar[list[BindingType]] = [
        Binding("ctrl+t", "switch_theme"),
    ]

    def __init__(
        self, indexer: Indexer, use_cache: bool, search_query: str | None
    ) -> None:
        super().__init__()
        self.indexer: Indexer = indexer
        self.use_cache: bool = use_cache
        self.search_query: str | None = search_query

        # load theme from config file
        theme = get_config().get("general.theme", "textual-dark")
        if theme not in self.available_themes:
            error_message = (
                f"invalid theme '{theme}' configured.\n"
                f"available themes: {', '.join(sorted(self.available_themes))}"
            )
            raise RuntimeError(error_message)
        self.theme = theme

    async def on_mount(self) -> None:
        if not (self.search_query and self.search_query.strip()):
            self._show_welcome_and_search()
        else:  # direct show search screen
            await self.push_screen(
                HomeScreen(
                    indexer=self.indexer,
                    search_query=self.search_query,
                    use_cache=self.use_cache,
                )
            )

    def action_switch_theme(self) -> None:
        self.push_screen(ThemeSelectorScreen())

    @work(exclusive=True)
    async def _show_welcome_and_search(self) -> None:
        if search_query := await self.push_screen_wait(
            WelcomeScreen(indexer=self.indexer)
        ):  # show both screens
            await self.push_screen(
                HomeScreen(
                    indexer=self.indexer,
                    search_query=search_query,
                    use_cache=self.use_cache,
                )
            )



from abc import ABC, abstractmethod
from typing import Any

from torrra._types import Torrent


class BaseIndexer(ABC):
    def __init__(self, url: str, api_key: str, timeout: int = 10):
        self.url: str = url.rstrip("/")
        self.api_key: str = api_key
        self.timeout: int = timeout

    @abstractmethod
    def get_search_url(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def get_healthcheck_url(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    async def search(self, query: str, use_cache: bool = True) -> list[Torrent]:
        raise NotImplementedError()

    @abstractmethod
    async def healthcheck(self) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def _normalize_result(self, r: dict[str, Any]) -> Torrent:
        raise NotImplementedError()



from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.screen import Screen
from textual.widgets import ContentSwitcher
from typing_extensions import override

from torrra._types import Indexer, TorrentStatus
from torrra.core.download import get_download_manager
from torrra.core.torrent import get_torrent_manager
from torrra.widgets.downloads import DownloadsContent
from torrra.widgets.search import SearchContent
from torrra.widgets.sidebar import Sidebar


class HomeScreen(Screen[None]):
    def __init__(self, indexer: Indexer, search_query: str, use_cache: bool):
        super().__init__()
        self.indexer: Indexer = indexer
        self.search_query: str = search_query
        self.use_cache: bool = use_cache

        self._sidebar: Sidebar
        self._content_switcher: ContentSwitcher
        self._downloads_content: DownloadsContent

    @override
    def compose(self) -> ComposeResult:
        with Horizontal(id="main_layout"):
            yield Sidebar(id="sidebar")
            with ContentSwitcher(initial="search_content", id="content_switcher"):
                yield DownloadsContent()
                yield SearchContent(
                    indexer=self.indexer,
                    search_query=self.search_query,
                    use_cache=self.use_cache,
                )

    def on_mount(self) -> None:
        self._sidebar = self.query_one(Sidebar)
        self._sidebar.can_focus = True  # re-enable focus

        self._content_switcher = self.query_one(ContentSwitcher)
        self._downloads_content = self.query_one(DownloadsContent)

        # start torrents in background
        tm, dm = get_torrent_manager(), get_download_manager()
        torrents = tm.get_all_torrents()
        for torrent in torrents:
            dm.add_torrent(
                torrent["magnet_uri"],
                is_paused=torrent["is_paused"],
            )

        # start timer to update data on both sidebar
        # and downloads content table
        self.set_interval(1, self._update_downloads_data)

    def on_sidebar_item_selected(self, event: Sidebar.ItemSelected) -> None:
        self.query_one(ContentSwitcher).current = event.group_id

    def on_search_content_download_requested(self) -> None:
        self.query_one(ContentSwitcher).current = "downloads_content"
        self.query_one(Sidebar).select_node_by_group_id("downloads_content")

        self._downloads_content.focus_table()

    def _update_downloads_data(self) -> None:
        dm = get_download_manager()
        magnet_uris = list(dm.torrents.keys())

        counts = {"Downloading": 0, "Seeding": 0, "Paused": 0, "Completed": 0}
        statuses: dict[str, TorrentStatus | None] = {}

        for uri in magnet_uris:
            status = dm.get_torrent_status(uri)
            statuses[uri] = status
            if not status:
                continue

            state_text = dm.get_torrent_state_text(status)
            if state_text in ("Downloading", "Fetching"):
                counts["Downloading"] += 1
            elif state_text in counts:
                counts[state_text] += 1

        self._sidebar.update_download_counts(counts)
        # only update downloads table if it is visible
        if self._content_switcher.current == "downloads_content":
            self._downloads_content.update_table_data(statuses)



from textual.app import ComposeResult
from textual.containers import Vertical
from textual.message import Message
from textual.widgets import ProgressBar, Static
from typing_extensions import override


class DetailsPanel(Vertical):
    class Closed(Message):
        """Posted when the panel is closed."""

    def __init__(self, show_progress_bar: bool = False) -> None:
        super().__init__(classes="hidden")
        self.show_progress_bar: bool = show_progress_bar
        # UI refs
        self._content_widget: Static
        self._progress_bar: ProgressBar | None = None

    @override
    def compose(self) -> ComposeResult:
        yield Static()
        if self.show_progress_bar:
            yield ProgressBar(total=100)

    def on_mount(self) -> None:
        self._content_widget = self.query_one(Static)
        if self.show_progress_bar:
            self._progress_bar = self.query_one(ProgressBar)
        # enable focus for this widget
        self.can_focus: bool = True

    def key_escape(self) -> None:
        self.add_class("hidden")
        self.post_message(self.Closed())

    def update_content(self, content: str, progress: float | None = None) -> None:
        self._content_widget.update(content)
        if self._progress_bar and progress is not None:
            self._progress_bar.progress = progress



from textual.app import ComposeResult
from textual.containers import Container, Grid
from textual.screen import Screen
from textual.widgets import Input, Static
from typing_extensions import override

from torrra._types import Indexer
from torrra._version import __version__

BANNER = """
▀█▀ █▀█ █▀▄ █▀▄ █▀▄ █▀█
 █  █ █ █▀▄ █▀▄ █▀▄ █▀█
 ▀  ▀▀▀ ▀ ▀ ▀ ▀ ▀ ▀ ▀ ▀
"""


class WelcomeScreen(Screen[str]):
    def __init__(self, indexer: Indexer) -> None:
        super().__init__()
        self.indexer: Indexer = indexer

    @override
    def compose(self) -> ComposeResult:
        with Container(id="welcome_container"):
            yield Static(BANNER, id="banner")
            yield Static(
                "\n".join(
                    [
                        "Find and download torrents right from here.",
                        "Powered by libtorrent and Python ❤️",
                    ]
                ),
                id="subtitle",
            )
            yield Input(placeholder="Search...", id="search")
            yield Static(
                f"v{__version__}{f' - {self.indexer.name}' if self.indexer else ''}",
                id="version",
            )
            with Container(id="commands_container"):
                with Grid():
                    yield Static("[key binds]", id="title", markup=False)
                    yield Static("[q]uit", markup=False)
                    yield Static("ctrl+q", classes="key")
                    yield Static("[t]heme switcher", markup=False)
                    yield Static("ctrl+t", classes="key")

    async def on_input_submitted(self, event: Input.Submitted) -> None:
        if query := event.value.strip():
            self.dismiss(query)



import sqlite3
from contextlib import contextmanager
from pathlib import Path

from platformdirs import user_data_dir

DB_DIR = Path(user_data_dir("torrra"))
DB_FILE = DB_DIR / "torrra.db"


@contextmanager
def get_db_connection():
    DB_DIR.mkdir(parents=True, exist_ok=True)
    conn = sqlite3.connect(DB_FILE)
    try:
        yield conn
    finally:
        conn.close()


def init_db() -> None:
    with get_db_connection() as conn:
        cursor = conn.cursor()
        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS torrents (
                magnet_uri TEXT PRIMARY KEY,
                title TEXT,
                size REAL,
                source TEXT,
                is_paused BOOLEAN DEFAULT 0
            )
            """
        )
        conn.commit()



def human_readable_size(size_bytes: float, short: bool = False) -> str:
    if not short:
        units = ["B", "KB", "MB", "GB", "TB"]
        for unit in units:
            if size_bytes < 1024.0:
                return f"{size_bytes:.2f} {unit}"
            size_bytes /= 1024.0
        return f"{size_bytes:.2f} PB"

    # short version
    if size_bytes < 1024:
        return f"{int(size_bytes)}B"

    for unit in ["K", "M", "G", "T"]:
        size_bytes /= 1024.0
        if size_bytes < 1024.0:
            number = (
                f"{size_bytes:.1f}".rstrip("0").rstrip(".")
                if size_bytes < 10
                else str(int(size_bytes))
            )
            return f"{number}{unit}"
    return f"{int(size_bytes)}P"


def lazy_import(dotted_path: str):
    import importlib

    try:
        module_path, obj_name = dotted_path.rsplit(".", 1)
        module = importlib.import_module(module_path)
        return getattr(module, obj_name)
    except (ModuleNotFoundError, AttributeError) as e:
        raise ImportError(f"failed to import: {dotted_path}\n{e}")



import sqlite3
from functools import lru_cache

from torrra._types import Torrent, TorrentRecord
from torrra.core.db import get_db_connection, init_db


@lru_cache
def get_torrent_manager() -> "TorrentManager":
    init_db()
    return TorrentManager()


class TorrentManager:
    def add_torrent(self, torrent: Torrent) -> None:
        with get_db_connection() as conn:
            cursor = conn.cursor()
            cursor.execute(
                """
                INSERT OR IGNORE INTO torrents (magnet_uri, title, size, source)
                VALUES (?, ?, ?, ?)
                """,
                (
                    torrent.magnet_uri,
                    torrent.title,
                    torrent.size,
                    torrent.source,
                ),
            )
            conn.commit()

    def remove_torrent(self, magnet_uri: str) -> None:
        with get_db_connection() as conn:
            cursor = conn.cursor()
            cursor.execute("DELETE FROM torrents WHERE magnet_uri = ?", (magnet_uri,))
            conn.commit()

    def update_torrent_paused_state(self, magnet_uri: str, is_paused: bool) -> None:
        with get_db_connection() as conn:
            cursor = conn.cursor()
            cursor.execute(
                "UPDATE torrents SET is_paused = ? WHERE magnet_uri = ?",
                (int(is_paused), magnet_uri),
            )
            conn.commit()

    def get_all_torrents(self) -> list[TorrentRecord]:
        with get_db_connection() as conn:
            conn.row_factory = sqlite3.Row
            cursor = conn.cursor()
            cursor.execute("SELECT * FROM torrents")
            rows = cursor.fetchall()

            return [
                TorrentRecord(
                    magnet_uri=row["magnet_uri"],
                    title=row["title"],
                    size=row["size"],
                    source=row["source"],
                    is_paused=bool(row["is_paused"]),
                )
                for row in rows
            ]



from dataclasses import asdict, dataclass
from typing import Any, Literal, TypedDict

import libtorrent as lt


# TORRENT TYPES
class TorrentDict(TypedDict):
    """Dict variant of Torrent dataclass."""

    magnet_uri: str
    title: str
    size: float
    seeders: int
    leechers: int
    source: str


class TorrentStatus(TypedDict):
    """Torrent status on upload and download."""

    state: lt.torrent_status.states
    progress: float
    down_speed: float
    up_speed: float
    seeders: int
    leechers: int
    is_paused: bool


class TorrentRecord(TypedDict):
    """Torrent data stored in db."""

    magnet_uri: str
    title: str
    size: float
    source: str
    is_paused: bool


@dataclass
class Torrent:
    """Torrent I/O dataclass."""

    magnet_uri: str
    title: str
    size: float
    seeders: int
    leechers: int
    source: str

    @classmethod
    def from_dict(cls, d: TorrentDict) -> "Torrent":
        return cls(**d)

    def to_dict(self) -> dict[str, Any]:
        return asdict(self)


# INDEXER TYPES
IndexerName = Literal["jackett", "prowlarr"]


@dataclass
class Indexer:
    """Indexer dataclass."""

    name: IndexerName
    url: str
    api_key: str



from contextlib import suppress
from typing import cast

from textual import on, work
from textual.app import ComposeResult
from textual.containers import Vertical
from textual.message import Message
from textual.widgets import Input, Static
from typing_extensions import override

from torrra._types import Indexer, Torrent
from torrra.core.torrent import get_torrent_manager
from torrra.indexers.base import BaseIndexer
from torrra.utils.helpers import human_readable_size, lazy_import
from torrra.utils.magnet import resolve_magnet_uri
from torrra.widgets.data_table import AutoResizingDataTable
from torrra.widgets.details_panel import DetailsPanel
from torrra.widgets.spinner import Spinner


class SearchContent(Vertical):
    COLS: list[tuple[str, str, int]] = [
        ("No", "no_col", 2),
        ("Title", "title_col", 25),
        ("Size", "size_col", 10),
        ("S:L", "seeders_leechers_col", 6),
    ]

    class SearchResults(Message):
        def __init__(self, results: list[Torrent], query: str) -> None:
            self.results: list[Torrent] = results
            self.query: str = query
            super().__init__()

    class DownloadRequested(Message):
        def __init__(self, torrent: Torrent) -> None:
            self.torrent: Torrent = torrent
            super().__init__()

    def __init__(self, indexer: Indexer, search_query: str, use_cache: bool):
        super().__init__(id="search_content")
        self.indexer: Indexer = indexer
        self.search_query: str = search_query
        self.use_cache: bool = use_cache

        # instance-level cache
        self._indexer_instance_cache: BaseIndexer | None = None

        # application states
        self._search_results_map: dict[str, Torrent] = {}
        self._selected_torrent: Torrent | None = None

        # ui refs (cached later)
        self._search_input: Input
        self._table: AutoResizingDataTable[str]
        self._details_panel: DetailsPanel
        self._loader: Vertical

    @override
    def compose(self) -> ComposeResult:
        yield Input(placeholder="Search...", value=self.search_query)
        yield AutoResizingDataTable(cursor_type="row", classes="hidden")
        yield DetailsPanel()
        with Vertical(id="loader"):
            yield Static()
            yield Spinner(name="shark")

    def on_mount(self) -> None:
        self._search_input = self.query_one(Input)
        self._search_input.border_title = "search"
        self._search_input.focus()

        self._table = self.query_one(AutoResizingDataTable)
        self._table.expand_col = "title_col"
        self._table.border_title = "results"

        self._details_panel = self.query_one(DetailsPanel)
        self._details_panel.border_title = "details"

        self._loader = self.query_one("#loader", Vertical)
        # setup table
        for label, key, width in self.COLS:
            self._table.add_column(label, width=width, key=key)
        # send initial search
        self.post_message(Input.Submitted(self._search_input, self.search_query))

    async def key_enter(self) -> None:
        if not self._details_panel.has_focus:
            return

        if self._selected_torrent:
            # uri returned from the indexer
            # might not be a proper magnet uri, resolve anyways
            raw_magnet_uri = self._selected_torrent.magnet_uri
            resolved_magnet_uri = await resolve_magnet_uri(raw_magnet_uri)

            if resolved_magnet_uri is None:
                return

            # update with resolved magnet_uri
            self._selected_torrent.magnet_uri = resolved_magnet_uri

            tm = get_torrent_manager()
            tm.add_torrent(self._selected_torrent)
            self.post_message(self.DownloadRequested(self._selected_torrent))

    def on_input_submitted(self, event: Input.Submitted) -> None:
        query = event.value
        if not query:
            return

        self._table.add_class("hidden")
        self._table.clear()
        self._details_panel.add_class("hidden")
        self._loader.remove_class("hidden")
        cast(Spinner, self._loader.children[1]).resume()
        cast(Static, self._loader.children[0]).update(
            f"Searching for [b]{query}[/b]..."
        )

        self._perform_search(query)

    @work(exclusive=True)
    async def _perform_search(self, query: str) -> None:
        indexer = self._get_indexer_instance()
        results = []

        with suppress(Exception):
            results = await indexer.search(query, use_cache=self.use_cache)
        self.post_message(self.SearchResults(results, query))

    @on(SearchResults)
    def on_search_results(self, message: SearchResults) -> None:
        if not message.results:
            cast(Spinner, self._loader.children[1]).pause()
            cast(Static, self._loader.children[0]).update(
                f"Nothing Found for [b]{message.query}[/b]"
            )  # show loader and exit
            return

        self._loader.add_class("hidden")
        self._table.remove_class("hidden")
        self._table.focus()  # initial focus table
        self._table.border_title = (
            f"{self._table.border_title} ({len(message.results)})"
        )

        seen: set[str] = set()
        for idx, torrent in enumerate(message.results):
            if torrent.magnet_uri in seen:
                continue

            seen.add(torrent.magnet_uri)
            self._search_results_map[torrent.magnet_uri] = torrent
            self._table.add_row(
                str(idx + 1),
                torrent.title,
                human_readable_size(torrent.size),
                f"{str(torrent.seeders)}:{str(torrent.leechers)}",
                key=torrent.magnet_uri,
            )

    def on_details_panel_closed(self):
        self._selected_torrent = None
        self._table.focus()

    def on_data_table_row_selected(
        self, event: AutoResizingDataTable.RowSelected
    ) -> None:
        magnet_uri = cast(str, event.row_key.value)
        self._selected_torrent = self._search_results_map.get(magnet_uri)
        if not self._selected_torrent:
            return

        details = f"""
[b]{self._selected_torrent.title}[/b]
[b]Size:[/b] {human_readable_size(self._selected_torrent.size)} - [b]Seeders:[/b] {self._selected_torrent.seeders} - [b]Leechers:[/b] {self._selected_torrent.leechers} - [b]Source:[/b] {self._selected_torrent.source}

[dim]Press 'enter' to download or 'esc' to close.[/dim]
"""

        self._details_panel.update_content(details.strip())
        self._details_panel.remove_class("hidden")
        self._details_panel.focus()

    def _get_indexer_instance(self) -> BaseIndexer:
        if self._indexer_instance_cache:
            return self._indexer_instance_cache

        name = self.indexer.name
        indexer_cls_str = f"torrra.indexers.{name}.{name.title()}Indexer"

        indexer_cls = lazy_import(indexer_cls_str)
        assert issubclass(indexer_cls, BaseIndexer)
        indexer_instance = indexer_cls(
            url=self.indexer.url, api_key=self.indexer.api_key
        )

        self._indexer_instance_cache = indexer_instance
        return indexer_instance



from typing import cast

from textual.app import ComposeResult
from textual.containers import Vertical
from typing_extensions import override

from torrra._types import TorrentRecord, TorrentStatus
from torrra.core.download import DownloadManager, get_download_manager
from torrra.core.torrent import get_torrent_manager
from torrra.utils.helpers import human_readable_size
from torrra.widgets.data_table import AutoResizingDataTable
from torrra.widgets.details_panel import DetailsPanel


class DownloadsContent(Vertical):
    COLS: list[tuple[str, str, int]] = [
        ("No", "no_col", 2),
        ("Title", "title", 25),
        ("St.", "status", 3),
        ("Done", "done_percent", 4),
        ("Up", "up_speed", 6),
        ("Down", "down_speed", 6),
    ]

    def __init__(self) -> None:
        super().__init__(id="downloads_content")
        self._torrents: list[TorrentRecord] = []
        self._selected_torrent: TorrentRecord | None = None

        self._dm: DownloadManager = get_download_manager()

        self._table: AutoResizingDataTable[str]
        self._details_panel: DetailsPanel

    @override
    def compose(self) -> ComposeResult:
        yield AutoResizingDataTable(cursor_type="row")
        yield DetailsPanel(show_progress_bar=True)

    def on_mount(self) -> None:
        self._table = self.query_one(AutoResizingDataTable)
        self._table.expand_col = "title"

        self._details_panel = self.query_one(DetailsPanel)
        self._details_panel.border_title = "details"
        # setup table
        for label, key, width in self.COLS:
            self._table.add_column(label, width=width, key=key)

    def on_show(self) -> None:
        tm = get_torrent_manager()
        self._torrents = tm.get_all_torrents()

        self._table.clear()
        self._table.border_title = f"all ({len(self._torrents)})"

        for idx, torrent in enumerate(self._torrents):
            self._dm.add_torrent(torrent["magnet_uri"], is_paused=torrent["is_paused"])
            self._table.add_row(
                str(idx + 1),
                torrent["title"],
                "N/A",
                "0%",
                "0 B/s",
                "0 B/s",
                key=torrent["magnet_uri"],
            )

    def key_p(self) -> None:
        if not self._selected_torrent:
            return

        magnet_uri = self._selected_torrent["magnet_uri"]
        self._dm.toggle_pause(magnet_uri)
        if status := self._dm.get_torrent_status(magnet_uri):
            tm = get_torrent_manager()
            tm.update_torrent_paused_state(magnet_uri, status["is_paused"])

    def key_d(self) -> None:
        if not self._selected_torrent:
            return

        magnet_uri = self._selected_torrent["magnet_uri"]
        self._dm.remove_torrent(magnet_uri)

        tm = get_torrent_manager()
        tm.remove_torrent(magnet_uri)

        self._table.remove_row(magnet_uri)
        self._torrents = [t for t in self._torrents if t["magnet_uri"] != magnet_uri]
        self._table.border_title = f"all ({len(self._torrents)})"
        self._details_panel.add_class("hidden")
        self._selected_torrent = None

    def on_details_panel_closed(self):
        self._selected_torrent = None

    def on_data_table_row_selected(
        self, event: AutoResizingDataTable.RowSelected
    ) -> None:
        row_key = cast(str, event.row_key.value)
        self._selected_torrent = next(
            (d for d in self._torrents if d["magnet_uri"] == row_key), None
        )

        if self._selected_torrent:
            if status := self._dm.get_torrent_status(
                self._selected_torrent["magnet_uri"]
            ):
                self._update_details_panel(status)
            self._details_panel.remove_class("hidden")
            self._details_panel.focus()
        else:  # selected torrent is invalid
            self._details_panel.add_class("hidden")

    def focus_table(self) -> None:
        self._table.focus()

    def update_table_data(self, statuses: dict[str, TorrentStatus | None]) -> None:
        if not self._torrents:
            return

        for torrent in self._torrents:
            status = statuses.get(torrent["magnet_uri"])
            if not status:
                continue

            self._table.update_cell(
                torrent["magnet_uri"],
                "status",
                self._dm.get_torrent_state_text(status, short=True),
            )
            self._table.update_cell(
                torrent["magnet_uri"],
                "done_percent",
                f"{int(status['progress'])}%",
            )
            self._table.update_cell(
                torrent["magnet_uri"],
                "up_speed",
                f"{human_readable_size(status['up_speed'], short=True)}/s",
            )
            self._table.update_cell(
                torrent["magnet_uri"],
                "down_speed",
                f"{human_readable_size(status['down_speed'], short=True)}/s",
            )

            if (
                self._selected_torrent
                and self._selected_torrent["magnet_uri"] == torrent["magnet_uri"]
            ):
                # update the details panel if its open and
                # showing this torrent data
                self._update_details_panel(status)

    def _update_details_panel(self, status: TorrentStatus) -> None:
        if not self._selected_torrent:
            return

        state_text = self._dm.get_torrent_state_text(status)
        size = human_readable_size(float(self._selected_torrent["size"]))
        up_speed = f"{human_readable_size(status['up_speed'])}/s"
        down_speed = f"{human_readable_size(status['down_speed'])}/s"

        details = f"""
[b]{self._selected_torrent["title"]}[/b]
[b]Size:[/b] {size} - [b]Status:[/b] {state_text} - [b]Source:[/b] {self._selected_torrent["source"]}
[b]S/L:[/b] {status["seeders"]}/{status["leechers"]} - [b]Up:[/b] {up_speed} - [b]Down:[/b] {down_speed}

[dim]Press 'p' to pause/resume, 'd' to delete, or 'esc' to close.[/dim]
"""
        # update details panel internal widgets
        self._details_panel.update_content(
            details.strip(),
            progress=status["progress"],
        )



import asyncio
from typing import Any

from textual import work
from textual.app import ComposeResult
from textual.binding import Binding, BindingType
from textual.containers import Vertical
from textual.screen import ModalScreen
from textual.widgets import Label, ListItem, ListView, Static
from textual.worker import Worker
from typing_extensions import override

from torrra.core.config import get_config


class ThemeSelectorScreen(ModalScreen[None]):
    BINDINGS: list[BindingType] = [
        Binding("escape", "close_screen"),
        Binding("k", "cursor_up"),
        Binding("j", "cursor_down"),
    ]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.original_theme: str = self.app.theme

        self._list_view: ListView | None = None
        self._update_worker: Worker[None] | None = None

    @override
    def compose(self) -> ComposeResult:
        with Vertical(id="theme-switcher-container"):
            yield Label("[b]Select a Theme[/b]")
            yield Label("j/k: navigate - enter: save", markup=False)
            yield ListView(
                *[
                    ListItem(Static(theme), name=theme)
                    for theme in sorted(self.app.available_themes)
                ]
            )

    def on_mount(self) -> None:
        self._list_view = self.query_one(ListView)
        for i, item in enumerate(self._list_view.children):
            if item.name == self.app.theme:
                self._list_view.index = i
                break

    def on_list_view_highlighted(self, event: ListView.Highlighted) -> None:
        self._cancel_update_worker()
        if event.item is not None and event.item.name is not None:
            self._update_worker = self._set_theme(event.item.name)

    def on_list_view_selected(self) -> None:
        self._cancel_update_worker()
        if self.app.theme != self.original_theme:
            get_config().set("general.theme", self.app.theme)
        self.app.pop_screen()

    def action_close_screen(self) -> None:
        self._cancel_update_worker()
        self.app.theme = self.original_theme
        self.app.pop_screen()

    def action_cursor_up(self) -> None:
        if self._list_view is not None:
            self._list_view.action_cursor_up()

    def action_cursor_down(self) -> None:
        if self._list_view is not None:
            self._list_view.action_cursor_down()

    @work(exclusive=True)
    async def _set_theme(self, theme: str) -> None:
        await asyncio.sleep(0.5)
        self.app.theme = theme

    def _cancel_update_worker(self) -> None:
        if self._update_worker is not None:
            self._update_worker.cancel()



from functools import lru_cache

import libtorrent as lt

from torrra._types import TorrentStatus
from torrra.core.config import get_config


@lru_cache
def get_download_manager() -> "DownloadManager":
    return DownloadManager()


class DownloadManager:
    _STATE_MAP: dict[lt.torrent_status.states, tuple[str, str]] = {
        lt.torrent_status.states.downloading: ("Downloading", "DL"),
        lt.torrent_status.states.seeding: ("Seeding", "SE"),
        lt.torrent_status.states.finished: ("Completed", "CD"),
        lt.torrent_status.states.downloading_metadata: ("Fetching", "FE"),
    }

    def __init__(self) -> None:
        self.session: lt.session = lt.session({"listen_interfaces": "0.0.0.0:6881"})
        self.torrents: dict[str, lt.torrent_handle] = {}

    def add_torrent(self, magnet_uri: str, is_paused: bool = False) -> None:
        if magnet_uri in self.torrents:
            return

        params = {"save_path": get_config().get("general.download_path")}
        if is_paused:
            params["flags"] = lt.torrent_flags.paused
        # start tracking provided magnet_uri
        self.torrents[magnet_uri] = lt.add_magnet_uri(self.session, magnet_uri, params)

    def remove_torrent(self, magnet_uri: str) -> None:
        handle = self.torrents.get(magnet_uri)
        if handle and handle.is_valid():
            self.session.remove_torrent(handle)
            del self.torrents[magnet_uri]

    def toggle_pause(self, magnet_uri: str) -> None:
        handle = self.torrents.get(magnet_uri)
        if not handle or not handle.is_valid():
            return

        status = handle.status()
        if (status.flags & lt.torrent_flags.paused) != 0:
            handle.resume()
        else:  # if not paused
            handle.pause()

    def get_torrent_status(self, magnet_uri: str) -> TorrentStatus | None:
        handle = self.torrents.get(magnet_uri)
        if not handle or not handle.is_valid():
            return None

        s = handle.status()
        return TorrentStatus(
            state=s.state,
            progress=s.progress * 100,
            down_speed=s.download_rate,
            up_speed=s.upload_rate,
            seeders=s.num_seeds,
            leechers=s.num_peers,
            is_paused=(s.flags & lt.torrent_flags.paused) != 0,
        )

    def get_torrent_state_text(self, status: TorrentStatus, short: bool = False) -> str:
        if status["is_paused"]:
            return "Paused" if not short else "PD"

        idx = 1 if short else 0
        return self._STATE_MAP.get(status["state"], ("N/A", "N/A"))[idx]



import time
from typing import Any, ClassVar, TypeVar

from textual.binding import Binding, BindingType
from textual.reactive import reactive
from textual.widgets import DataTable
from textual.widgets.data_table import ColumnKey

T = TypeVar("T")


class AutoResizingDataTable(DataTable[T]):
    expand_col: reactive[str | None] = reactive(None)

    BINDINGS: ClassVar[list[BindingType]] = [
        Binding("k", "cursor_up"),
        Binding("j", "cursor_down"),
        Binding("G", "scroll_bottom"),
        Binding("ctrl+u", "page_up"),
        Binding("ctrl+d", "page_down"),
        Binding("l", "select_cursor"),
    ]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._last_g_press: float = 0

    def key_g(self) -> None:
        current_time = time.time()
        if current_time - self._last_g_press < 0.4:
            self.action_scroll_top()
            self._last_g_press = 0
        else:  # save for next event
            self._last_g_press = current_time

    def on_resize(self) -> None:
        self._resize_columns()
        self.refresh()

    def _resize_columns(self) -> None:
        if not self.columns or not self.expand_col:
            return

        total_cell_padding = self.cell_padding * 2 * len(self.columns)
        expand_col_key = ColumnKey(self.expand_col)

        other_cols_width = sum(
            col.width for key, col in self.columns.items() if key != expand_col_key
        )

        available_width = self.size.width - total_cell_padding
        expand_col_width = available_width - other_cols_width

        if expand_col_width > 0:
            self.columns[expand_col_key].width = expand_col_width



from typing import Any, ClassVar

from textual.app import ComposeResult
from textual.message import Message
from textual.reactive import Reactive, reactive
from textual.widgets import Tree
from textual.widgets.tree import TreeNode
from typing_extensions import override


class Sidebar(Tree[Any]):
    DOWNLOADS_GROUP: ClassVar[list[str]] = [
        "Downloading",
        "Seeding",
        "Paused",
        "Completed",
    ]

    class ItemSelected(Message):
        def __init__(self, group_id: str, group_type: str | None = None) -> None:
            self.group_id: str = group_id
            self.group_type: str | None = group_type
            super().__init__()

    def __init__(self, id: str) -> None:
        super().__init__("Menu", id=id)
        self.show_horizontal_scrollbar: Reactive[bool] = reactive(False)
        self.show_root: bool = False
        self.guide_depth: int = 3
        self.can_focus: bool = False  # re-enable focus later

        self._downloads_root_node: TreeNode[Any]
        self._downloads_nodes: dict[str, TreeNode[Any]] = {}

    @override
    def compose(self) -> ComposeResult:
        root = self.root
        # NODES
        search = root.add("Search", allow_expand=False)
        search.data = {"group_id": "search_content"}

        downloads_node = root.add("Downloads (0)", expand=True, allow_expand=False)
        downloads_node.data = {"group_id": "downloads_content"}

        for item in self.DOWNLOADS_GROUP:
            node = downloads_node.add(f"{item} (0)", allow_expand=False)
            node.data = {**downloads_node.data, "group_type": item}

            self._downloads_nodes[item] = node
        self._downloads_root_node = downloads_node

        self.select_node(search)  # default
        return super().compose()

    def on_tree_node_selected(self, event: Tree.NodeSelected[dict[str, str]]) -> None:
        node_data = event.node.data or {}
        if group_id := node_data.get("group_id"):
            self.post_message(
                self.ItemSelected(
                    group_id=group_id, group_type=node_data.get("group_type")
                )
            )

    def select_node_by_group_id(self, group_id: str) -> None:
        def _find_node(node: TreeNode[Any]) -> TreeNode[Any] | None:
            if node.data and node.data.get("group_id") == group_id:
                return node

            for child in node.children:
                if found := _find_node(child):
                    return found
            return None

        if target_node := _find_node(self.root):
            self.select_node(target_node)

    def update_download_counts(self, counts: dict[str, int]) -> None:
        total = sum(counts.values())
        self._downloads_root_node.set_label(f"Downloads ({total})")

        for item in self.DOWNLOADS_GROUP:
            count = counts.get(item, 0)
            self._downloads_nodes[item].set_label(f"{item} ({count})")



from typing import Any, cast

import httpx
from typing_extensions import override

from torrra._types import Torrent, TorrentDict
from torrra.core.cache import cache
from torrra.core.exceptions import IndexerError
from torrra.indexers.base import BaseIndexer


class ProwlarrIndexer(BaseIndexer):
    @override
    def get_search_url(self) -> str:
        return f"{self.url}/api/v1/search"

    @override
    def get_healthcheck_url(self) -> str:
        return f"{self.url}/api/v1/health"

    @override
    async def search(self, query: str, use_cache: bool = True) -> list[Torrent]:
        key = cache.make_key("prowlarr", query)

        if use_cache and key in cache:
            raw_data = cast(list[TorrentDict], cache.get(key))
            return [Torrent.from_dict(d) for d in raw_data]

        url = self.get_search_url()
        params = {"apikey": self.api_key, "query": query}

        async with httpx.AsyncClient(timeout=self.timeout) as client:
            resp = await client.get(url, params=params)
            resp.raise_for_status()
            torrents = [self._normalize_result(r) for r in resp.json()]

        if use_cache:
            cache.set(key, [t.to_dict() for t in torrents])

        return torrents

    @override
    async def healthcheck(self) -> bool:
        url = self.get_healthcheck_url()
        params = {"apikey": self.api_key}

        async with httpx.AsyncClient(timeout=self.timeout) as client:
            try:
                resp = await client.get(url, params=params)
                resp.raise_for_status()
                return True

            except httpx.RequestError:
                raise IndexerError(
                    "could not connect to prowlarr server\n"
                    + "please make sure prowlarr server is running and the url is correct"
                )

            except httpx.HTTPStatusError as e:
                status_code = e.response.status_code

                if status_code == 401:
                    raise IndexerError(
                        "invalid prowlarr server api key\n"
                        + "double-check the api key you provided"
                    )
                else:
                    raise IndexerError(
                        f"prowlarr server returned http {status_code}\n"
                        + "unexpected response from prowlarr server. please verify your setup"
                    )

    @override
    def _normalize_result(self, r: dict[str, Any]) -> Torrent:
        return Torrent(
            title=r.get("title", "unknown"),
            size=r.get("size", 0),
            seeders=r.get("seeders", 0),
            leechers=r.get("leechers", 0),
            source=r.get("indexer", "unknown"),
            magnet_uri=r.get("magnetUrl") or r["downloadUrl"],
        )



import os
import tempfile
from contextlib import suppress

import httpx
import libtorrent as lt


async def resolve_magnet_uri(input_uri: str) -> str | None:
    if input_uri.startswith("magnet:"):
        return input_uri

    try:
        # some torrent doesnt have magnet uri, instead a uri to .torrent file
        # send request to that uri and parse content
        async with httpx.AsyncClient(follow_redirects=False) as client:
            resp = await client.get(input_uri)

        # get magnet uri from location header
        if resp.status_code in (301, 302):
            return resp.headers.get("location")

        # if its a .torrent file
        content_type = resp.headers.get("content-type", "")
        if "application/x-bittorrent" in content_type or input_uri.endswith(".torrent"):
            with tempfile.NamedTemporaryFile(suffix=".torrent", delete=False) as tmp:
                tmp.write(resp.content)
                tmp_path = tmp.name

            try:
                info = lt.torrent_info(tmp_path)
                return lt.make_magnet_uri(info)
            finally:
                # delete file silently
                with suppress(Exception):
                    os.remove(tmp_path)
    except Exception as e:
        print(f"an unexpected error occurred: {e}")
    return None



import asyncio
from typing import cast

import click

from torrra._types import Indexer, IndexerName
from torrra.core.config import get_config
from torrra.core.exceptions import ConfigError, IndexerError
from torrra.indexers.base import BaseIndexer
from torrra.utils.helpers import lazy_import


def run_with_indexer(
    *,
    name: IndexerName,
    indexer_cls_str: str,
    url: str | None,
    api_key: str | None,
    no_cache: bool,
    search_query: str | None = None,
) -> None:
    config = get_config()
    # validate command args
    if (url is None) != (api_key is None):
        click.secho(
            "both --url and --api-key must be provided together, or neither to use config.",
            fg="red",
            err=True,
        )
        return

    # url, api_key config fallback
    if url is None and api_key is None:
        try:
            url = config.get(f"indexers.{name}.url")
            api_key = config.get(f"indexers.{name}.api_key")
        except ConfigError as e:
            click.secho(f"{e}\ncheck your configuration file.", fg="red", err=True)
            return

    click.secho(f"connecting to {name} server at {url}", fg="cyan")

    # import indexers only when needed
    indexer_cls = lazy_import(indexer_cls_str)

    # type narrowing for pyright
    assert url is not None and api_key is not None

    # async indexer validation
    async def healthcheck_indexer():
        try:
            assert issubclass(indexer_cls, BaseIndexer)
            return await indexer_cls(url, api_key).healthcheck()
        except IndexerError as e:
            click.secho(str(e), fg="red", err=True)
            return False

    if not asyncio.run(healthcheck_indexer()):
        return

    # update/store indexers configuration
    config.set(f"indexers.{name}.url", url)
    config.set(f"indexers.{name}.api_key", api_key)

    # load app only when needed (heavy stuff)
    from torrra.app import TorrraApp

    try:
        # determine cache settings with CLI override
        use_cache = config.get("general.use_cache", True)
        if no_cache:  # --no-cache flag overrides config
            use_cache = False

        indexer = Indexer(name, url, api_key)
        app = TorrraApp(indexer, use_cache=use_cache, search_query=search_query)
        app.run()
    except RuntimeError as e:
        click.secho(str(e), fg="red", err=True)


def run_with_default_indexer(
    *, no_cache: bool, search_query: str | None = None
) -> None:
    config = get_config()
    try:
        default_indexer = config.get("indexers.default")
        if not default_indexer:
            raise ConfigError("no default indexer specified under [indexers.default].")

        url = config.get(f"indexers.{default_indexer}.url")
        api_key = config.get(f"indexers.{default_indexer}.api_key")

        run_with_indexer(
            name=cast(IndexerName, default_indexer),
            indexer_cls_str=f"torrra.indexers.{default_indexer}.{default_indexer.title()}Indexer",
            url=url,
            api_key=api_key,
            no_cache=no_cache,
            search_query=search_query,
        )
    except ConfigError as e:
        click.secho(f"{e}\ncheck your configuration file.", fg="red", err=True)



import time
from typing import Any

from rich.spinner import Spinner as _Spinner
from textual.timer import Timer
from textual.widgets import Static


class Spinner(Static):
    def __init__(self, name: str, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._spinner: _Spinner = _Spinner(name)
        self.update_timer: Timer

    def on_mount(self) -> None:
        self.update_timer = self.set_interval(1 / 15, self.update_spinner, pause=False)

    def update_spinner(self) -> None:
        self.update(self._spinner.render(time.monotonic()))

    def resume(self) -> None:
        self.update_timer.resume()
        self.update_spinner()

    def pause(self) -> None:
        self.update_timer.pause()
        self.update_spinner()



import atexit
import hashlib
from typing import Any

from diskcache import Cache as _Cache
from platformdirs import user_cache_dir
from typing_extensions import override

from torrra.core.config import get_config
from torrra.core.constants import DEFAULT_CACHE_TTL


class Cache(_Cache):
    @override
    def set(
        self,
        key: str,
        value: Any,
        expire: int | None = None,
        read: bool = False,
        tag: Any = None,
        retry: bool = False,
    ):
        expire = get_config().get("general.cache_ttl", DEFAULT_CACHE_TTL)
        return super().set(key, value, expire, read, tag, retry)

    def make_key(self, prefix: str, query: str) -> str:
        return f"{prefix}:{hashlib.sha256(query.encode()).hexdigest()}"


cache = Cache(user_cache_dir("torrra"))
atexit.register(cache.close)



import ast
from contextlib import suppress
from functools import lru_cache
from pathlib import Path
from typing import Any, cast

import tomli_w
import tomllib
from platformdirs import user_config_dir, user_downloads_dir

from torrra.core.constants import DEFAULT_CACHE_TTL
from torrra.core.exceptions import ConfigError

CONFIG_DIR = Path(user_config_dir("torrra"))
CONFIG_FILE = CONFIG_DIR / "config.toml"

# sentinel value used for robust
# config.get(..., default=...) value check
_sentinel = object()


@lru_cache
def get_config() -> "Config":
    return Config()


class Config:
    def __init__(self) -> None:
        self.config: dict[str, Any] = {}
        self._load_config()

    def get(self, key_path: str, default: Any | None = _sentinel) -> Any:
        keys = key_path.split(".")
        current = self.config

        try:
            for key in keys:
                current = current[key]

            if isinstance(current, dict):
                raise ConfigError(
                    f"key does not contain a value (it's a section): {key_path}"
                )
            return current

        except (KeyError, TypeError):
            if default is not _sentinel:
                return default

            if len(keys) > 1:
                raise ConfigError(f"key does not contain a section: {key_path}")
            raise ConfigError(f"key not found: {key_path}")

    def set(self, key_path: str, value: str) -> None:
        current = self.config
        keys = key_path.split(".")

        try:
            for key in keys[:-1]:
                if key not in current:
                    current[key] = {}
                elif not isinstance(current[key], dict):
                    raise ConfigError(
                        f"cannot set '{key_path}': '{key}' is not a section"
                    )
                current = current[key]

            new_value: Any = value
            # handle case-insensitive "true"/"false" for booleans
            if value.lower() == "true":
                new_value = True
            elif value.lower() == "false":
                new_value = False
            # handle other literals (int, float, etc.)
            else:  # convert data type silently
                with suppress(ValueError, SyntaxError):
                    new_value = ast.literal_eval(value)

            current[keys[-1]] = new_value
            self._save_config()

        except (KeyError, TypeError) as e:
            raise ConfigError(f"failed to set '{key_path}': {str(e)}")

    def list(self) -> list[str]:
        results: list[str] = []

        def _flatten_config(data: dict[str, Any], prefix: str = "") -> None:
            # recursively iterate through config
            for key, value in data.items():
                # construct new prefix
                new_prefix = f"{prefix}.{key}" if prefix else key
                if isinstance(value, dict):
                    # if dict, recurse deeper
                    _flatten_config(cast(dict[str, Any], value), new_prefix)
                else:
                    if isinstance(value, bool):
                        # if bool, convert to lowercase string
                        value = str(value).lower()
                    # append flattened key-value pair
                    results.append(f"{new_prefix}={value}")

        _flatten_config(self.config)
        return results

    def _load_config(self) -> None:
        if not CONFIG_FILE.exists():
            self._create_default_config()
            self._save_config()

        try:
            with open(CONFIG_FILE, "rb") as f:
                self.config = tomllib.load(f)
        except Exception as e:
            print(f"loading config failed: {e}")

    def _create_default_config(self) -> None:
        self.config = {
            "general": {
                "download_path": user_downloads_dir(),
                "download_in_external_client": False,
                "theme": "textual-dark",
                "use_cache": True,
                "cache_ttl": DEFAULT_CACHE_TTL,
            }
        }

    def _save_config(self) -> None:
        CONFIG_DIR.mkdir(parents=True, exist_ok=True)
        with open(CONFIG_FILE, "wb") as f:
            tomli_w.dump(self.config, f)
